use std::sync::Arc;

use actix::Addr;
use actix_web::{web, HttpResponse, Responder};
use askama::Template;

use crate::actors::bruteforce_actor::BruteForceActor;
use crate::actors::providers_states_actor::{ProviderLoginState, ProvidersStatesActor};
use crate::actors::{bruteforce_actor, providers_states_actor};
use crate::constants::{APP_NAME, MAX_FAILED_LOGIN_ATTEMPTS};
use crate::controllers::base_controller::{build_fatal_error_page, redirect_user};
use crate::controllers::login_controller::BaseLoginPage;
use crate::data::action_logger::{Action, ActionLogger};
use crate::data::login_redirect::LoginRedirect;
use crate::data::provider::{ProviderID, ProvidersManager};
use crate::data::provider_configuration::ProviderConfigurationHelper;
use crate::data::remote_ip::RemoteIP;

#[derive(askama::Template)]
#[template(path = "login/prov_login_error.html")]
struct ProviderLoginError<'a> {
    _p: BaseLoginPage<'a>,
    message: &'a str,
}

impl<'a> ProviderLoginError<'a> {
    pub fn get(message: &'a str, redirect_uri: &'a LoginRedirect) -> HttpResponse {
        let body = Self {
            _p: BaseLoginPage {
                danger: None,
                success: None,
                page_title: "Upstream login",
                app_name: APP_NAME,
                redirect_uri,
            },
            message,
        }
        .render()
        .unwrap();

        HttpResponse::Unauthorized()
            .content_type("text/html")
            .body(body)
    }
}

#[derive(serde::Deserialize)]
pub struct StartLoginQuery {
    #[serde(default)]
    redirect: LoginRedirect,
    id: ProviderID,
}

/// Start user authentication using a provider
pub async fn start_login(
    remote_ip: RemoteIP,
    providers: web::Data<Arc<ProvidersManager>>,
    states: web::Data<Addr<ProvidersStatesActor>>,
    query: web::Query<StartLoginQuery>,
    logger: ActionLogger,
) -> impl Responder {
    // Get provider information
    let provider = match providers.find_by_id(&query.id) {
        None => {
            return HttpResponse::NotFound()
                .body(build_fatal_error_page("Login provider not found!"));
        }
        Some(p) => p,
    };

    // Generate & save state
    let state = ProviderLoginState::new(&provider.id, query.redirect.clone());
    states
        .send(providers_states_actor::RecordState {
            ip: remote_ip.0,
            state: state.clone(),
        })
        .await
        .unwrap();

    logger.log(Action::StartLoginAttemptWithOpenIDProvider {
        provider_id: &provider.id,
        state: &state.state_id,
    });

    // Get provider configuration
    let config = match ProviderConfigurationHelper::get_configuration(&provider).await {
        Ok(c) => c,
        Err(e) => {
            log::error!("Failed to load provider configuration! {}", e);
            return HttpResponse::InternalServerError().body(build_fatal_error_page(
                "Failed to load provider configuration!",
            ));
        }
    };

    log::debug!("Provider configuration: {:?}", config);

    let url = config.auth_url(&provider, &state);
    log::debug!("Redirect user on {url} for authentication",);

    // Redirect user
    redirect_user(&url)
}

#[derive(serde::Deserialize)]
pub struct FinishLoginSuccess {
    code: String,
    state: String,
}

#[derive(serde::Deserialize)]
pub struct FinishLoginError {
    error: String,
    error_description: Option<String>,
}

#[derive(serde::Deserialize)]
pub struct FinishLoginQuery {
    #[serde(flatten)]
    success: Option<FinishLoginSuccess>,
    #[serde(flatten)]
    error: Option<FinishLoginError>,
}

/// Finish user authentication using a provider
pub async fn finish_login(
    remote_ip: RemoteIP,
    providers: web::Data<Arc<ProvidersManager>>,
    states: web::Data<Addr<ProvidersStatesActor>>,
    bruteforce: web::Data<Addr<BruteForceActor>>,
    query: web::Query<FinishLoginQuery>,
    logger: ActionLogger,
) -> impl Responder {
    let query = match query.0.success {
        Some(q) => q,
        None => {
            let error_message = query
                .0
                .error
                .map(|e| e.error_description.unwrap_or(e.error))
                .unwrap_or("Authentication failed (unspecified error)!".to_string());

            logger.log(Action::ProviderError {
                message: error_message.as_str(),
            });

            return ProviderLoginError::get(&error_message, &LoginRedirect::default());
        }
    };

    // Get & consume state
    let state = states
        .send(providers_states_actor::ConsumeState {
            ip: remote_ip.0,
            state_id: query.state.clone(),
        })
        .await
        .unwrap();

    let state = match state {
        Some(s) => s,
        None => {
            logger.log(Action::ProviderCBInvalidState {
                state: query.state.as_str(),
            });
            log::warn!("User returned invalid state!");
            return ProviderLoginError::get("Invalid state!", &LoginRedirect::default());
        }
    };

    // We perform rate limiting before attempting to use authorization code
    let failed_attempts = bruteforce
        .send(bruteforce_actor::CountFailedAttempt {
            ip: remote_ip.into(),
        })
        .await
        .unwrap();

    if failed_attempts > MAX_FAILED_LOGIN_ATTEMPTS {
        logger.log(Action::ProviderRateLimited);
        return HttpResponse::TooManyRequests().body(build_fatal_error_page(
            "Too many failed login attempts, please try again later!",
        ));
    }

    // Retrieve provider information & configuration
    let provider = providers
        .find_by_id(&state.provider_id)
        .expect("Unable to retrieve provider information!");

    let provider_config = match ProviderConfigurationHelper::get_configuration(&provider).await {
        Ok(c) => c,
        Err(e) => {
            log::error!("Failed to load provider configuration! {}", e);
            return HttpResponse::InternalServerError().body(build_fatal_error_page(
                "Failed to load provider configuration!",
            ));
        }
    };

    // Get access token & user information
    let token = provider_config.get_token(&provider, &query.code).await;
    let token = match token {
        Ok(t) => t,
        Err(e) => {
            log::error!("Failed to retrieve login token! {:?}", e);

            bruteforce
                .send(bruteforce_actor::RecordFailedAttempt {
                    ip: remote_ip.into(),
                })
                .await
                .unwrap();

            logger.log(Action::ProviderFailedGetToken {
                state: &state,
                code: query.code.as_str(),
            });

            return ProviderLoginError::get(
                "Failed to retrieve login token from identity provider!",
                &state.redirect,
            );
        }
    };

    // Use access token to get user information
    let info = provider_config.get_userinfo(&token).await;
    println!("info: {:?}", info);

    // TODO : check if user is authorized to access application
    // TODO : check if 2FA is enabled
    // TODO : redirect user to login route
    // TODO : add proper logging
    HttpResponse::Ok().body("continue")
}
